@@ -17,22 +17,23 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
 */
 
-#include <vector>
-#include <fstream>
-
 #include <stdlib.h>
+#include <string.h>
+#include <fstream>
+#include <vector>
 
-#include "osal_preproc.h"
-#include "m64p_types.h"
-#include "m64p_plugin.h"
-#include "m64p_config.h"
-
+#define M64P_PLUGIN_PROTOTYPES 1
 #include "Config.h"
 #include "Debugger.h"
 #include "DeviceBuilder.h"
 #include "RenderBase.h"
+#include "Texture.h"
 #include "TextureManager.h"
 #include "Video.h"
+#include "m64p_config.h"
+#include "m64p_plugin.h"
+#include "m64p_types.h"
+#include "osal_preproc.h"
 
 #define INI_FILE        "RiceVideoLinux.ini"
 
@@ -55,41 +56,6 @@ const char *frameBufferSettings[] =
 "With Emulator Write Only",
 };
 
-const int resolutions[][2] =
-{
-{320, 240},
-{400, 300},
-{480, 360},
-{512, 384},
-{640, 480},
-{800, 600},
-{1024, 768},
-{1152, 864},
-{1280, 960},
-{1400, 1050},
-{1600, 1200},
-{1920, 1440},
-{2048, 1536},
-};
-const int numberOfResolutions = sizeof(resolutions)/sizeof(int)/2;
-
-const char* resolutionsS[] =
-{
-"320 x 240",
-"400 x 300",
-"480 x 360",
-"512 x 384",
-"640 x 480",
-"800 x 600",
-"1024 x 768",
-"1152 x 864",
-"1280 x 960",
-"1400 x 1050",
-"1600 x 1200",
-"1920 x 1440",
-"2048 x 1536"
-};
-
 const char *frameBufferWriteBackControlSettings[] =
 {
 "Every Frame (default)",
@@ -180,18 +146,6 @@ SettingInfo openGLDepthBufferSettings[] =
 {"32-bit", 32},
 };
 
-RenderEngineSetting OpenGLRenderSettings[] =
-{
-{"To Fit Your Video Card", OGL_DEVICE},
-{"OpenGL 1.1 (Lowest)",  OGL_1_1_DEVICE},
-{"OpenGL 1.2/1.3", OGL_1_2_DEVICE},
-{"OpenGL 1.4", OGL_1_4_DEVICE},
-//{"OpenGL 1.4, the 2nd combiner",  OGL_1_4_V2_DEVICE},
-{"OpenGL for Nvidia TNT or better", OGL_TNT2_DEVICE},
-{"OpenGL for Nvidia GeForce or better ", NVIDIA_OGL_DEVICE},
-{"OpenGL Fragment Program Extension", OGL_FRAGMENT_PROGRAM},
-};
-
 SettingInfo OnScreenDisplaySettings[] =
 {
 {"Display Nothing", ONSCREEN_DISPLAY_NOTHING},
@@ -204,8 +158,6 @@ SettingInfo OnScreenDisplaySettings[] =
 {"Display Debug Information With Core Msgs", ONSCREEN_DISPLAY_DEBUG_INFORMATION_WITH_CORE_MSG},
 };
 
-const int numberOfOpenGLRenderEngineSettings = sizeof(OpenGLRenderSettings)/sizeof(RenderEngineSetting);
-
 void GenerateFrameBufferOptions(void)
 {
     if( CDeviceBuilder::GetGeneralDeviceType() == OGL_DEVICE )
@@ -217,76 +169,76 @@ void GenerateFrameBufferOptions(void)
             currentRomOptions.N64RenderToTextureEmuType = TXT_BUF_IGNORE;
     }
 
-    frameBufferOptions.bUpdateCIInfo            = false;
+    frameBufferOptions.bUpdateCIInfo                       = false;
 
-    frameBufferOptions.bCheckBackBufs           = false;
-    frameBufferOptions.bWriteBackBufToRDRAM     = false;
-    frameBufferOptions.bLoadBackBufFromRDRAM    = false;
+    frameBufferOptions.bCheckBackBufs                      = false;
+    frameBufferOptions.bWriteBackBufToRDRAM                = false;
+    frameBufferOptions.bLoadBackBufFromRDRAM               = false;
 
-    frameBufferOptions.bIgnore                  = true;
+    frameBufferOptions.bIgnore                             = true;
 
-    frameBufferOptions.bSupportRenderTextures           = false;
-    frameBufferOptions.bCheckRenderTextures         = false;
-    frameBufferOptions.bRenderTextureWriteBack          = false;
-    frameBufferOptions.bLoadRDRAMIntoRenderTexture      = false;
+    frameBufferOptions.bSupportRenderTextures              = false;
+    frameBufferOptions.bCheckRenderTextures                = false;
+    frameBufferOptions.bRenderTextureWriteBack             = false;
+    frameBufferOptions.bLoadRDRAMIntoRenderTexture         = false;
 
-    frameBufferOptions.bProcessCPUWrite         = false;
-    frameBufferOptions.bProcessCPURead          = false;
-    frameBufferOptions.bAtEachFrameUpdate       = false;
-    frameBufferOptions.bIgnoreRenderTextureIfHeightUnknown      = false;
+    frameBufferOptions.bProcessCPUWrite                    = false;
+    frameBufferOptions.bProcessCPURead                     = false;
+    frameBufferOptions.bAtEachFrameUpdate                  = false;
+    frameBufferOptions.bIgnoreRenderTextureIfHeightUnknown = false;
 
     switch( currentRomOptions.N64FrameBufferEmuType )
     {
     case FRM_BUF_NONE:
         break;
     case FRM_BUF_COMPLETE:
-        frameBufferOptions.bAtEachFrameUpdate       = true;
-        frameBufferOptions.bProcessCPUWrite         = true;
-        frameBufferOptions.bProcessCPURead          = true;
-        frameBufferOptions.bUpdateCIInfo            = true;
+        frameBufferOptions.bAtEachFrameUpdate    = true;
+        frameBufferOptions.bProcessCPUWrite      = true;
+        frameBufferOptions.bProcessCPURead       = true;
+        frameBufferOptions.bUpdateCIInfo         = true;
         break;
     case FRM_BUF_WRITEBACK_AND_RELOAD:
-        frameBufferOptions.bLoadBackBufFromRDRAM    = true;
+        frameBufferOptions.bLoadBackBufFromRDRAM = true;
     case FRM_BUF_BASIC_AND_WRITEBACK:
-        frameBufferOptions.bWriteBackBufToRDRAM     = true;
+        frameBufferOptions.bWriteBackBufToRDRAM  = true;
     case FRM_BUF_BASIC:
-        frameBufferOptions.bCheckBackBufs           = true;
+        frameBufferOptions.bCheckBackBufs        = true;
     case FRM_BUF_IGNORE:
-        frameBufferOptions.bUpdateCIInfo            = true;
+        frameBufferOptions.bUpdateCIInfo         = true;
         break;
     case FRM_BUF_BASIC_AND_WITH_EMULATOR:
         // Banjo Kazooie
-        frameBufferOptions.bCheckBackBufs           = true;
+        frameBufferOptions.bCheckBackBufs        = true;
     case FRM_BUF_WITH_EMULATOR:
-        frameBufferOptions.bUpdateCIInfo            = true;
-        frameBufferOptions.bProcessCPUWrite         = true;
-        frameBufferOptions.bProcessCPURead          = true;
+        frameBufferOptions.bUpdateCIInfo         = true;
+        frameBufferOptions.bProcessCPUWrite      = true;
+        frameBufferOptions.bProcessCPURead       = true;
         break;
     case FRM_BUF_WITH_EMULATOR_READ_ONLY:
-        frameBufferOptions.bUpdateCIInfo            = true;
-        frameBufferOptions.bProcessCPURead          = true;
+        frameBufferOptions.bUpdateCIInfo         = true;
+        frameBufferOptions.bProcessCPURead       = true;
         break;
     case FRM_BUF_WITH_EMULATOR_WRITE_ONLY:
-        frameBufferOptions.bUpdateCIInfo            = true;
-        frameBufferOptions.bProcessCPUWrite         = true;
+        frameBufferOptions.bUpdateCIInfo         = true;
+        frameBufferOptions.bProcessCPUWrite      = true;
         break;
     }
 
     switch( currentRomOptions.N64RenderToTextureEmuType )
     {
     case TXT_BUF_NONE:
-        frameBufferOptions.bSupportRenderTextures           = false;
+        frameBufferOptions.bSupportRenderTextures      = false;
         break;
     case TXT_BUF_WRITE_BACK_AND_RELOAD:
-        frameBufferOptions.bLoadRDRAMIntoRenderTexture      = true;
+        frameBufferOptions.bLoadRDRAMIntoRenderTexture = true;
     case TXT_BUF_WRITE_BACK:
-        frameBufferOptions.bRenderTextureWriteBack          = true;
+        frameBufferOptions.bRenderTextureWriteBack     = true;
     case TXT_BUF_NORMAL:
-        frameBufferOptions.bCheckRenderTextures         = true;
-        frameBufferOptions.bIgnore                  = false;
+        frameBufferOptions.bCheckRenderTextures        = true;
+        frameBufferOptions.bIgnore                     = false;
     case TXT_BUF_IGNORE:
-        frameBufferOptions.bUpdateCIInfo            = true;
-        frameBufferOptions.bSupportRenderTextures           = true;
+        frameBufferOptions.bUpdateCIInfo               = true;
+        frameBufferOptions.bSupportRenderTextures      = true;
         break;
     }
 
@@ -308,15 +260,60 @@ BOOL InitConfiguration(void)
         DebugMessage(M64MSG_ERROR, "Unable to open Video-Rice configuration section");
         return FALSE;
     }
+    
+    /* Ensure versionning is present */
+    int ConfigParamsVersion;
+    if (ConfigGetParameter(l_ConfigVideoRice, "Version", M64TYPE_INT, &ConfigParamsVersion, sizeof(int)) != M64ERR_SUCCESS)
+    {
+        DebugMessage(M64MSG_WARNING, "No version number in 'Rice-Video' config section. Setting defaults.");
+        ConfigParamsVersion = 0; // zero to ensure updates will be applied
+    }
+    /* Update config parameters */
+    if (ConfigParamsVersion < CONFIG_PARAM_VERSION)
+    {
+        DebugMessage(M64MSG_WARNING, "Old parameter config version detected : %d, updating to %d;", ConfigParamsVersion, CONFIG_PARAM_VERSION);
+        if (ConfigParamsVersion == 0) /* From v0 to v1: Remove OGL_TNT2_DEVICE and NVIDIA_OGL device, Remove force fog method option */
+        {
+            int oldOglDevice;
+            if (ConfigGetParameter(l_ConfigVideoRice, "OpenGLRenderSetting", M64TYPE_INT, &oldOglDevice, sizeof(int)) == M64ERR_SUCCESS)
+            {
+                /* OGL_1.1 was 1, OGL_1.2 was 2, OGL_1.3 was 3, OGL_1.4 was 4, OGL_1.4_V2 was 5, OGL_TNT2_DEVICE was 6, NVIDIA_OGL was 7 but doesnt exist anymore: Put to auto*/
+                if ((oldOglDevice == 1) || (oldOglDevice == 2) || (oldOglDevice == 3) || (oldOglDevice == 4) || (oldOglDevice == 5) || (oldOglDevice == 6) || (oldOglDevice == 7))
+                {
+                    oldOglDevice = 0; // auto
+                }
+                else if (oldOglDevice >= 8) /* OGL_FRAGMENT_PROGRAM (was 8+) is now 1*/
+                {
+                    oldOglDevice = 1;
+                }
+                ConfigSetParameter(l_ConfigVideoRice, "OpenGLRenderSetting", M64TYPE_INT, &oldOglDevice);
+                ConfigSetParameterHelp(l_ConfigVideoRice, "OpenGLRenderSetting", "OpenGL level to support (0=auto, 1=OGL_FRAGMENT_PROGRAM)");
+            }
+            int fogMethod;
+            if (ConfigGetParameter(l_ConfigVideoRice, "FogMethod", M64TYPE_INT, &fogMethod, sizeof(int)) == M64ERR_SUCCESS)
+            {
+                if ( fogMethod > 1 ) // if FogMethod was "Force Fog"
+                {
+                    fogMethod = 1;
+                    ConfigSetParameter(l_ConfigVideoRice, "FogMethod", M64TYPE_INT, &fogMethod);
+                }
+                ConfigSetParameterHelp(l_ConfigVideoRice, "FogMethod", "Enable, Disable fog generation (0=Disable, 1=Enable)");
+            }
+            ConfigParamsVersion = 1;
+        } // place others update stuff here and increment "ConfigParamsVersion" and CONFIG_PARAM_VERSION each time.
+    }
 
-    ConfigSetDefaultBool(l_ConfigVideoGeneral, "Fullscreen", 0, "Use fullscreen mode if True, or windowed mode if False ");
+    /* Set default config parameters */
+    ConfigSetDefaultBool(l_ConfigVideoGeneral, "Fullscreen", 0, "Use fullscreen mode if True, or windowed mode if False");
     ConfigSetDefaultInt(l_ConfigVideoGeneral, "ScreenWidth", 640, "Width of output window or fullscreen width");
     ConfigSetDefaultInt(l_ConfigVideoGeneral, "ScreenHeight", 480, "Height of output window or fullscreen height");
+    ConfigSetDefaultBool(l_ConfigVideoGeneral, "VerticalSync", 0, "If true, activate the SDL_GL_SWAP_CONTROL attribute");
 
+    ConfigSetDefaultInt(l_ConfigVideoRice, "Version", CONFIG_PARAM_VERSION, "Mupen64Plus Rice Video Plugin config parameter version number");
     ConfigSetDefaultInt(l_ConfigVideoRice, "FrameBufferSetting", FRM_BUF_NONE, "Frame Buffer Emulation (0=ROM default, 1=disable)");
     ConfigSetDefaultInt(l_ConfigVideoRice, "FrameBufferWriteBackControl", FRM_BUF_WRITEBACK_NORMAL, "Frequency to write back the frame buffer (0=every frame, 1=every other frame, etc)");
     ConfigSetDefaultInt(l_ConfigVideoRice, "RenderToTexture", TXT_BUF_NONE, "Render-to-texture emulation (0=none, 1=ignore, 2=normal, 3=write back, 4=write back and reload)");
-#if defined(WIN32)
+#if defined(WIN32) || defined(__APPLE__)
     ConfigSetDefaultInt(l_ConfigVideoRice, "ScreenUpdateSetting", SCREEN_UPDATE_AT_1ST_CI_CHANGE, "Control when the screen will be updated (0=ROM default, 1=VI origin update, 2=VI origin change, 3=CI change, 4=first CI change, 5=first primitive draw, 6=before screen clear, 7=after screen drawn)");  // SCREEN_UPDATE_AT_VI_UPDATE_AND_DRAWN
 #else
     ConfigSetDefaultInt(l_ConfigVideoRice, "ScreenUpdateSetting", SCREEN_UPDATE_AT_VI_UPDATE, "Control when the screen will be updated (0=ROM default, 1=VI origin update, 2=VI origin change, 3=CI change, 4=first CI change, 5=first primitive draw, 6=before screen clear, 7=after screen drawn)");  // SCREEN_UPDATE_AT_VI_UPDATE_AND_DRAWN
@@ -334,26 +331,29 @@ BOOL InitConfiguration(void)
     ConfigSetDefaultBool(l_ConfigVideoRice, "FullTMEMEmulation", FALSE, "N64 Texture Memory Full Emulation (may fix some games, may break others)");
     ConfigSetDefaultBool(l_ConfigVideoRice, "OpenGLVertexClipper", FALSE, "Enable vertex clipper for fog operations");
     ConfigSetDefaultBool(l_ConfigVideoRice, "EnableSSE", TRUE, "Enable/Disable SSE optimizations for capable CPUs");
-    ConfigSetDefaultBool(l_ConfigVideoRice, "EnableVertexShader", FALSE, "Use GPU vertex shader");
     ConfigSetDefaultBool(l_ConfigVideoRice, "SkipFrame", FALSE, "If this option is enabled, the plugin will skip every other frame");
     ConfigSetDefaultBool(l_ConfigVideoRice, "TexRectOnly", FALSE, "If enabled, texture enhancement will be done only for TxtRect ucode");
     ConfigSetDefaultBool(l_ConfigVideoRice, "SmallTextureOnly", FALSE, "If enabled, texture enhancement will be done only for textures width+height<=128");
+    ConfigSetDefaultBool(l_ConfigVideoRice, "LoadHiResCRCOnly", TRUE, "Select hi-resolution textures based only on the CRC and ignore format+size information (Glide64 compatibility)");
     ConfigSetDefaultBool(l_ConfigVideoRice, "LoadHiResTextures", FALSE, "Enable hi-resolution texture file loading");
     ConfigSetDefaultBool(l_ConfigVideoRice, "DumpTexturesToFiles", FALSE, "Enable texture dumping");
     ConfigSetDefaultBool(l_ConfigVideoRice, "ShowFPS", FALSE, "Display On-screen FPS");
-    ConfigSetDefaultBool(l_ConfigVideoRice, "EnableMipmaping", TRUE, "Enable/Disable Mipmaping");
 
-    ConfigSetDefaultInt(l_ConfigVideoRice, "FogMethod", 0, "Enable, Disable or Force fog generation (0=Disable, 1=Enable n64 choose, 2=Force Fog)");
+    ConfigSetDefaultInt(l_ConfigVideoRice, "Mipmapping", 2, "Use Mipmapping? 0=no, 1=nearest, 2=bilinear, 3=trilinear");
+    ConfigSetDefaultInt(l_ConfigVideoRice, "FogMethod", 1, "Enable, Disable fog generation (0=Disable, 1=Enable)");
     ConfigSetDefaultInt(l_ConfigVideoRice, "ForceTextureFilter", 0, "Force to use texture filtering or not (0=auto: n64 choose, 1=force no filtering, 2=force filtering)");
-    ConfigSetDefaultInt(l_ConfigVideoRice, "TextureFilteringMethod", 1, "Choose wich texture filtering method will be used by your graphic card(0=no filtering, 1=bilinear, 2=trilinear)");
     ConfigSetDefaultInt(l_ConfigVideoRice, "TextureEnhancement", 0, "Primary texture enhancement filter (0=None, 1=2X, 2=2XSAI, 3=HQ2X, 4=LQ2X, 5=HQ4X, 6=Sharpen, 7=Sharpen More, 8=External, 9=Mirrored)");
     ConfigSetDefaultInt(l_ConfigVideoRice, "TextureEnhancementControl", 0, "Secondary texture enhancement filter (0 = none, 1-4 = filtered)");
     ConfigSetDefaultInt(l_ConfigVideoRice, "TextureQuality", TXT_QUALITY_DEFAULT, "Color bit depth to use for textures (0=default, 1=32 bits, 2=16 bits)");
     ConfigSetDefaultInt(l_ConfigVideoRice, "OpenGLDepthBufferSetting", 16, "Z-buffer depth (only 16 or 32)");
     ConfigSetDefaultInt(l_ConfigVideoRice, "MultiSampling", 0, "Enable/Disable MultiSampling (0=off, 2,4,8,16=quality)");
     ConfigSetDefaultInt(l_ConfigVideoRice, "ColorQuality", TEXTURE_FMT_A8R8G8B8, "Color bit depth for rendering window (0=32 bits, 1=16 bits)");
-    ConfigSetDefaultInt(l_ConfigVideoRice, "OpenGLRenderSetting", OGL_DEVICE, "OpenGL level to support (0=auto, 1=OGL_1.1, 2=OGL_1.2, 3=OGL_1.3, 4=OGL_1.4, 5=OGL_1.4_V2, 6=OGL_TNT2, 7=NVIDIA_OGL, 8=OGL_FRAGMENT_PROGRAM)");
-    ConfigSetDefaultInt(l_ConfigVideoRice, "AnisotropicFiltering", 0, "Enable/Disable Anisotropic Filtering for Mipmaping (0=no filtering, 2-16=quality). This is uneffective if EnableMipmaping is false. If the given value is to high to be supported by your graphic card, the value will be the highest value your graphic card can support. Better result with Trilinear filtering");
+    ConfigSetDefaultInt(l_ConfigVideoRice, "OpenGLRenderSetting", OGL_DEVICE, "OpenGL level to support (0=auto, 1=OGL_FRAGMENT_PROGRAM)");
+    ConfigSetDefaultInt(l_ConfigVideoRice, "AnisotropicFiltering", 0, "Enable/Disable Anisotropic Filtering for Mipmapping (0=no filtering, 2-16=quality). This is uneffective if Mipmapping is 0. If the given value is to high to be supported by your graphic card, the value will be the highest value your graphic card can support. Better result with Trilinear filtering");
+
+    ConfigSetDefaultBool(l_ConfigVideoRice, "ForcePolygonOffset", FALSE, "If true, use polygon offset values specified below");
+    ConfigSetDefaultFloat(l_ConfigVideoRice, "PolygonOffsetFactor", 0.0f, "Specifies a scale factor that is used to create a variable depth offset for each polygon");
+    ConfigSetDefaultFloat(l_ConfigVideoRice, "PolygonOffsetUnits", 0.0f, "Is multiplied by an implementation-specific value to create a constant depth offset");
     return TRUE;
 }
 
@@ -393,22 +393,23 @@ bool isSSESupported()
 {
     int SSESupport = 0;
 
-    // And finally, check the CPUID for Streaming SIMD Extensions support.
+// And finally, check the CPUID for Streaming SIMD Extensions support.
 #if !defined(__GNUC__) && !defined(NO_ASM)
-    _asm{
-       mov      eax, 1          // Put a "1" in eax to tell CPUID to get the feature bits
-         cpuid                  // Perform CPUID (puts processor feature info into EDX)
-         and        edx, 02000000h  // Test bit 25, for Streaming SIMD Extensions existence.
-         mov        SSESupport, edx // SIMD Extensions).  Set return value to 1 to indicate,
+    _asm
+	{
+            mov      eax, 1          // Put a "1" in eax to tell CPUID to get the feature bits
+            cpuid                    // Perform CPUID (puts processor feature info into EDX)
+            and      edx, 02000000h  // Test bit 25, for Streaming SIMD Extensions existence.
+            mov      SSESupport, edx // SIMD Extensions).  Set return value to 1 to indicate,
     }
 #elif defined(__GNUC__) && defined(__x86_64__) && !defined(NO_ASM)
   return true;
 #elif !defined(NO_ASM) // GCC assumed
    asm volatile (
          "push %%ebx                       \n"
-         "mov $1, %%eax                    \n"          // Put a "1" in eax to tell CPUID to get the feature bits
-         "cpuid                            \n"                  // Perform CPUID (puts processor feature info into EDX)
-         "and       $0x02000000, %%edx \n"  // Test bit 25, for Streaming SIMD Extensions existence.
+         "mov $1, %%eax                    \n"  // Put a "1" in eax to tell CPUID to get the feature bits
+         "cpuid                            \n"  // Perform CPUID (puts processor feature info into EDX)
+         "and       $0x02000000, %%edx     \n"  // Test bit 25, for Streaming SIMD Extensions existence.
          "pop %%ebx                        \n"
          : "=d"(SSESupport)
          :
@@ -427,6 +428,7 @@ static void ReadConfiguration(void)
     windowSetting.bDisplayFullscreen = ConfigGetParamBool(l_ConfigVideoGeneral, "Fullscreen");
     windowSetting.uDisplayWidth = ConfigGetParamInt(l_ConfigVideoGeneral, "ScreenWidth");
     windowSetting.uDisplayHeight = ConfigGetParamInt(l_ConfigVideoGeneral, "ScreenHeight");
+    windowSetting.bVerticalSync = ConfigGetParamBool(l_ConfigVideoGeneral, "VerticalSync");
 
     defaultRomOptions.N64FrameBufferEmuType = ConfigGetParamInt(l_ConfigVideoRice, "FrameBufferSetting");
     defaultRomOptions.N64FrameBufferWriteBackControl = ConfigGetParamInt(l_ConfigVideoRice, "FrameBufferWriteBackControl");
@@ -446,18 +448,17 @@ static void ReadConfiguration(void)
     options.bFullTMEM = ConfigGetParamBool(l_ConfigVideoRice, "FullTMEMEmulation");
     options.bOGLVertexClipper = ConfigGetParamBool(l_ConfigVideoRice, "OpenGLVertexClipper");
     options.bEnableSSE = ConfigGetParamBool(l_ConfigVideoRice, "EnableSSE");
-    options.bEnableVertexShader = ConfigGetParamBool(l_ConfigVideoRice, "EnableVertexShader");
     options.bSkipFrame = ConfigGetParamBool(l_ConfigVideoRice, "SkipFrame");
     options.bTexRectOnly = ConfigGetParamBool(l_ConfigVideoRice, "TexRectOnly");
     options.bSmallTextureOnly = ConfigGetParamBool(l_ConfigVideoRice, "SmallTextureOnly");
     options.bLoadHiResTextures = ConfigGetParamBool(l_ConfigVideoRice, "LoadHiResTextures");
+    options.bLoadHiResCRCOnly = ConfigGetParamBool(l_ConfigVideoRice, "LoadHiResCRCOnly");
     options.bDumpTexturesToFiles = ConfigGetParamBool(l_ConfigVideoRice, "DumpTexturesToFiles");
     options.bShowFPS = ConfigGetParamBool(l_ConfigVideoRice, "ShowFPS");
-    options.bEnableMipmaping = ConfigGetParamBool(l_ConfigVideoRice, "EnableMipmaping");
 
+    options.mipmapping = ConfigGetParamInt(l_ConfigVideoRice, "Mipmapping");
     options.fogMethod = ConfigGetParamInt(l_ConfigVideoRice, "FogMethod");
     options.forceTextureFilter = ConfigGetParamInt(l_ConfigVideoRice, "ForceTextureFilter");
-    options.textureFilteringMethod = ConfigGetParamInt(l_ConfigVideoRice, "TextureFilteringMethod");
     options.textureEnhancement = ConfigGetParamInt(l_ConfigVideoRice, "TextureEnhancement");
     options.textureEnhancementControl = ConfigGetParamInt(l_ConfigVideoRice, "TextureEnhancementControl");
     options.textureQuality = ConfigGetParamInt(l_ConfigVideoRice, "TextureQuality");
@@ -467,11 +468,14 @@ static void ReadConfiguration(void)
     options.OpenglRenderSetting = ConfigGetParamInt(l_ConfigVideoRice, "OpenGLRenderSetting");
     options.anisotropicFiltering = ConfigGetParamInt(l_ConfigVideoRice, "AnisotropicFiltering");
 
+    options.bForcePolygonOffset = ConfigGetParamBool(l_ConfigVideoRice, "ForcePolygonOffset");
+    options.polygonOffsetFactor = ConfigGetParamFloat(l_ConfigVideoRice, "PolygonOffsetFactor");
+    options.polygonOffsetUnits = ConfigGetParamFloat(l_ConfigVideoRice, "PolygonOffsetUnits");
+
     CDeviceBuilder::SelectDeviceType((SupportedDeviceType)options.OpenglRenderSetting);
 
     status.isMMXSupported = isMMXSupported();
     status.isSSESupported = isSSESupported();
-    status.isVertexShaderSupported = false;
 
     status.isSSEEnabled = status.isSSESupported && options.bEnableSSE;
 #if !defined(NO_ASM)
@@ -486,9 +490,6 @@ static void ReadConfiguration(void)
         ProcessVertexData = ProcessVertexDataNoSSE;
         DebugMessage(M64MSG_INFO, "Disabled SSE processing.");
     }
-
-    status.isVertexShaderEnabled = status.isVertexShaderSupported && options.bEnableVertexShader;
-    status.bUseHW_T_L = false;
 }
     
 BOOL LoadConfiguration(void)
@@ -673,16 +674,16 @@ void GenerateCurrentRomOptions()
 
     if( currentRomOptions.N64FrameBufferEmuType == 0 )      currentRomOptions.N64FrameBufferEmuType = defaultRomOptions.N64FrameBufferEmuType;
     else currentRomOptions.N64FrameBufferEmuType--;
-    if( currentRomOptions.N64RenderToTextureEmuType == 0 )      currentRomOptions.N64RenderToTextureEmuType = defaultRomOptions.N64RenderToTextureEmuType;
+    if( currentRomOptions.N64RenderToTextureEmuType == 0 )  currentRomOptions.N64RenderToTextureEmuType = defaultRomOptions.N64RenderToTextureEmuType;
     else currentRomOptions.N64RenderToTextureEmuType--;
     if( currentRomOptions.screenUpdateSetting == 0 )        currentRomOptions.screenUpdateSetting = defaultRomOptions.screenUpdateSetting;
     if( currentRomOptions.bNormalCombiner == 0 )            currentRomOptions.bNormalCombiner = defaultRomOptions.bNormalCombiner;
     else currentRomOptions.bNormalCombiner--;
-    if( currentRomOptions.bNormalBlender == 0 )         currentRomOptions.bNormalBlender = defaultRomOptions.bNormalBlender;
+    if( currentRomOptions.bNormalBlender == 0 )             currentRomOptions.bNormalBlender = defaultRomOptions.bNormalBlender;
     else currentRomOptions.bNormalBlender--;
     if( currentRomOptions.bFastTexCRC == 0 )                currentRomOptions.bFastTexCRC = defaultRomOptions.bFastTexCRC;
     else currentRomOptions.bFastTexCRC--;
-    if( currentRomOptions.bAccurateTextureMapping == 0 )        currentRomOptions.bAccurateTextureMapping = defaultRomOptions.bAccurateTextureMapping;
+    if( currentRomOptions.bAccurateTextureMapping == 0 )    currentRomOptions.bAccurateTextureMapping = defaultRomOptions.bAccurateTextureMapping;
     else currentRomOptions.bAccurateTextureMapping--;
 
     options.bUseFullTMEM = ((options.bFullTMEM && (g_curRomInfo.dwFullTMEM == 0)) || g_curRomInfo.dwFullTMEM == 2);
@@ -924,7 +925,7 @@ char * tidy(char * s)
     char * p = s + strlen(s);
 
     p--;
-    while (p >= s && (*p == ' ' || *p == 0xa || *p == '\n') )
+    while (p >= s && (*p == ' ' || *p == '\r' || *p == '\n') )
     {
         *p = 0;
         p--;
@@ -939,11 +940,19 @@ BOOL ReadIniFile()
     char readinfo[100];
     const char *ini_filepath = ConfigGetSharedDataFilepath(szIniFileName);
 
-    DebugMessage(M64MSG_VERBOSE, "Reading .ini file: %s", ini_filepath);
+    DebugMessage(M64MSG_VERBOSE, "Reading .ini file: %s", szIniFileName);
+
+    if (ini_filepath == NULL)
+    {
+        DebugMessage(M64MSG_ERROR, "Could not find .ini file: %s", szIniFileName);
+        return FALSE;
+    }
+
     inifile.open(ini_filepath);
 
     if (inifile.fail())
     {
+        DebugMessage(M64MSG_ERROR, "Could not open .ini file: %s", szIniFileName);
         return FALSE;
     }
 
@@ -954,7 +963,7 @@ BOOL ReadIniFile()
         if (readinfo[0] == '/')
             continue;
 
-        if (!strcasecmp(readinfo,"")==0)
+        if (strlen(readinfo) > 0)
         {
             if (readinfo[0] == '{') //if a section heading
             {
